#include "esp_camera.h"
#include <WiFi.h>
#include <driver/adc.h>
#include "FS.h"
#include "SD.h"
#include "SPI.h"
#include <Wire.h>
#include <DFRobot_DHT20.h>
#include <EEPROM.h>

//
// WARNING!!! PSRAM IC required for UXGA resolution and high JPEG quality
//            Ensure ESP32 Wrover Module or other board with PSRAM is selected
//            Partial images will be transmitted if image exceeds buffer size
//

// Select camera model
#define CAMERA_MODEL_AI_THINKER // Has PSRAM

#define PWDN_GPIO_NUM    -1
#define RESET_GPIO_NUM   12
#define XCLK_GPIO_NUM     4
#define SIOD_GPIO_NUM    17
#define SIOC_GPIO_NUM    18

#define Y9_GPIO_NUM      11
#define Y8_GPIO_NUM      10
#define Y7_GPIO_NUM       9
#define Y6_GPIO_NUM      46
#define Y5_GPIO_NUM       3
#define Y4_GPIO_NUM       8
#define Y3_GPIO_NUM      16
#define Y2_GPIO_NUM      15
#define VSYNC_GPIO_NUM    5
#define HREF_GPIO_NUM     6
#define PCLK_GPIO_NUM     7

#define	FREQ1			 1
#define	AN13			 13

#define LED_GREEN	42
#define LED_RED	 	41

#define LEDG_ON		digitalWrite( LED_GREEN, HIGH)
#define LEDG_OFF	digitalWrite( LED_GREEN, LOW)
#define LEDR_ON		digitalWrite( LED_RED, HIGH)
#define LEDR_OFF	digitalWrite( LED_RED, LOW)


const IPAddress 	subnet(255,255,255,0);

#define	RV8263_ADR	0x51

#define	VSPI_SCK	21
#define	VSPI_MOSI	47
#define	VSPI_MISO	14
#define	VSPI_SS		48
#define	SPI_FREQ	40000000

#define	I2C_SDA			1
#define	I2C_SCL			2	

float	temperature;
float	humidity;

int		time_ok;
int		wifi_ok;
int		eeprom_ok;
int		sdc_ok;
int		sd_save;

char	file_data[ 32768];


DFRobot_DHT20 dht20;

typedef	struct	{
	unsigned char 	year;
	unsigned char	month;
	unsigned char	date;
	unsigned char	week;
	unsigned char	hour;
	unsigned char	minute;
	unsigned char	sec;
} TIME_BUF;

typedef struct	{
	char			ssid[2][ 32];			// ssid
	char			pawd[2][ 32];			// password
	unsigned char	ip[2][ 4];				// ip address
	unsigned char	gway[ 2][ 4];			// gateway
	int				store_start;			// 0-23(start time)..24(off)
	int				store_end;				// 0-23(equal or over start time, other off)
	unsigned int	dummy;					// dummy 
	unsigned int	sum;					// check sum 
} PARA_BUF;

typedef struct	{
	int				file_num;
	char			file_name[ 10][ 16];	// 230728.csv...
	char			ref_name[ 4][ 16];		// 230728.csv...
	TIME_BUF		today;
	unsigned int	dummy;					// dummy 
	unsigned int	sum;					// check sum 
} FILE_BUF;


#define EEP_SIZE	1024

char	time_string[ 32];

TIME_BUF	now;
struct 	tm	tm_now;
PARA_BUF	para;
FILE_BUF	fname;

#define	MAX_DATA	1440	// 60*24 every minute data 1day

float	temp[ MAX_DATA];
float	humd[ MAX_DATA];
float	hist_temp[ MAX_DATA];
float	hist_humd[ MAX_DATA];
char	hist_name[ 16];
char	today_name[ 16];
int		data_num;
int		hist_data_num;

const char* ntpServer = "pool.ntp.org";
const long  gmtOffset_sec = 9*3600;
const int   daylightOffset_sec = 0;

void startCameraServer();
esp_err_t camera_capture( void);

int	 index_buf_len;

#if CONFIG_IDF_TARGET_ESP32S2 || CONFIG_IDF_TARGET_ESP32S3
#define VSPI FSPI
#endif

SPIClass * vspi = NULL;

unsigned int get_checksum( unsigned int *p, int size)
{
	int	 i;
	unsigned int	sum;
	
	sum = 0;
	for( i = 0 ; i < size; i++){
		sum += *p++;
	}
	return sum;
}

void	save_para( void)
{
	para.sum = get_checksum( (unsigned int*)&para, sizeof( para)/4-1);
	EEPROM.put<PARA_BUF>( 0, para);
	EEPROM.commit();
}

void	load_para( void)
{
	unsigned int	sum;
	int						 i;
	
	EEPROM.get<PARA_BUF>( 0, para);
	sum = get_checksum( (unsigned int*)&para, sizeof( para)/4-1);
	if( sum != para.sum){
		sprintf( para.ssid[ 0], "AirPort-----");
		sprintf( para.pawd[ 0], "-------------");
		para.ip[ 0][ 0] = 192;
		para.ip[ 0][ 1] = 168;
		para.ip[ 0][ 2] = 1;
		para.ip[ 0][ 3] = 35;
		para.gway[0][ 0] = 192;
		para.gway[0][ 1] = 168;
		para.gway[0][ 2] = 1;
		para.gway[0][ 3] = 1;
		sprintf( para.ssid[ 1], "AirPort-----");
		sprintf( para.pawd[ 1], "-------------");
		para.ip[ 1][ 0] = 192;
		para.ip[ 1][ 1] = 168;
		para.ip[ 1][ 2] = 0;
		para.ip[ 1][ 3] = 13;
		para.gway[ 1][ 0] = 192;
		para.gway[ 1][ 1] = 168;
		para.gway[ 1][ 2] = 0;
		para.gway[ 1][ 3] = 1;
		para.store_start	= 5*60;
		para.store_end		= 19*60;
		save_para();
	}
}

void	save_file_name( void)
{
	unsigned int	sum;
	int						 i;
	
	fname.sum = get_checksum( (unsigned int*)&fname, sizeof( fname)/4-1);
	EEPROM.put<FILE_BUF>( 512, fname);
	EEPROM.commit();
}

void	load_file_name( void)
{
	unsigned int	sum;
	int						 i;
	
	EEPROM.get<FILE_BUF>( 512, fname);
	sum = get_checksum( (unsigned int*)&fname, sizeof( fname)/4-1);
	if( sum != fname.sum){
		fname.file_num = 2;
		sprintf( fname.file_name[ 0], "230727.csv");
		sprintf( fname.file_name[ 1], "230728.csv");
		sprintf( fname.ref_name[0], "");
		sprintf( fname.ref_name[1], "");
		sprintf( fname.ref_name[2], "");
		sprintf( fname.ref_name[3], "");
		fname.today.date = 0;
		save_file_name();
	}
	
}
void	get_rv8263_data( TIME_BUF *p)
{
	unsigned char	b[ 7];
	
	Wire.beginTransmission( RV8263_ADR);
	Wire.write(4);
	Wire.endTransmission();
	Wire.requestFrom( RV8263_ADR, 7);
	while( Wire.available()<7);
	b[ 0] = Wire.read();			// sec
	b[ 1] = Wire.read();			// minute
	b[ 2] = Wire.read();			// hour
	b[ 3] = Wire.read();			// date
	b[ 4] = Wire.read();			// week
	b[ 5] = Wire.read();			// month
	b[ 6] = Wire.read();			// year
	p->sec 		= (b[ 0]/0x10)*10 + (b[ 0] & 0x0f);	// sec
	p->minute	= (b[ 1]/0x10)*10 + (b[ 1] & 0x0f);	// minute
	p->hour		= (b[ 2]/0x10)*10 + (b[ 2] & 0x0f);	// hour
	p->date		= (b[ 3]/0x10)*10 + (b[ 3] & 0x0f);	// date
	p->week		= 					(b[ 4] & 0x0f);	// week
	p->month	= (b[ 5]/0x10)*10 + (b[ 5] & 0x0f);	// month
	p->year		= (b[ 6]/0x10)*10 + (b[ 6] & 0x0f);	// year
}

void	set_rv8263_data( TIME_BUF d)
{
	unsigned char		b[ 7];
	
	b[ 0] = (d.sec /10)*0x10 	+ (d.sec % 10);		// sec
	b[ 1] = (d.minute /10)*0x10 + (d.minute % 10);	// minute
	b[ 2] = (d.hour  /10)*0x10 	+ (d.hour % 10);	// hour
	b[ 3] = (d.date  /10)*0x10 	+ (d.date % 10);	// date
	b[ 4] = 					  (d.week % 10);	// week
	b[ 5] = (d.month /10)*0x10	+ (d.month % 10);	// month
	b[ 6] = (d.year  /10)*0x10 	+ (d.year % 10);	// year
	Wire.beginTransmission( RV8263_ADR);
	Wire.write(4);
	Wire.write(b[ 0]);
	Wire.write(b[ 1]);
	Wire.write(b[ 2]);
	Wire.write(b[ 3]);
	Wire.write(b[ 4]);
	Wire.write(b[ 5]);
	Wire.write(b[ 6]);
	Wire.endTransmission();
}


void	rv8263_init( void)
{
	
}

void	set_freq( int f)
{
	ledcSetup( 0, 50, 10);		// 0ch 100Hz 12bit
	ledcAttachPin( FREQ1, 0);	// pin 0ch
  	ledcWrite( 0, 1024-f);		// 0ch pulse width
}

void	led_init( void)
{
	pinMode(LED_GREEN, OUTPUT);
	pinMode(LED_RED, OUTPUT);
	LEDR_ON;
	LEDG_OFF;
}

void listDir(fs::FS &fs, const char * dirname, uint8_t levels)
{
    Serial.printf("Listing directory: %s\n", dirname);

    File root = fs.open(dirname);
    if(!root){
        Serial.println("Failed to open directory");
        return;
    }
    if(!root.isDirectory()){
        Serial.println("Not a directory");
        return;
    }

    File file = root.openNextFile();
    while(file){
        if(file.isDirectory()){
            Serial.print("  DIR : ");
            Serial.println(file.name());
            if(levels){
                listDir(fs, file.path(), levels -1);
            }
        } else {
            Serial.print("  FILE: ");
            Serial.print(file.name());
            Serial.print("  SIZE: ");
            Serial.println(file.size());
        }
        file = root.openNextFile();
    }
}

void createDir(fs::FS &fs, const char * path)
{
    Serial.printf("Creating Dir: %s\n", path);
    if(fs.mkdir(path)){
        Serial.println("Dir created");
    } else {
        Serial.println("mkdir failed");
    }
}

void removeDir(fs::FS &fs, const char * path)
{
    Serial.printf("Removing Dir: %s\n", path);
    if(fs.rmdir(path)){
        Serial.println("Dir removed");
    } else {
        Serial.println("rmdir failed");
    }
}

int readFile(fs::FS &fs, const char * path)
{
    Serial.printf("Reading file: %s\n", path);

    File file = fs.open(path);
    if(!file){
        Serial.println("Failed to open file for reading");
        return 0;
    }

    Serial.print("Read from file: ");
    while(file.available()){
        Serial.write(file.read());
    }
    file.close();
    return 1;
}

void writeFile(fs::FS &fs, const char * path, const char * message)
{
    Serial.printf("Writing file: %s\n", path);

    File file = fs.open(path, FILE_WRITE);
    if(!file){
        Serial.println("Failed to open file for writing");
        return;
    }
    if(file.print(message)){
        Serial.println("File written");
    } else {
        Serial.println("Write failed");
    }
    file.close();
}

void appendFile(fs::FS &fs, const char * path, const char * message)
{
    Serial.printf("Appending to file: %s\n", path);

    File file = fs.open(path, FILE_APPEND);
    if(!file){
        Serial.println("Failed to open file for appending");
        return;
    }
    if(file.print(message)){
        Serial.println("Message appended");
    } else {
        Serial.println("Append failed");
    }
    file.close();
}

void renameFile(fs::FS &fs, const char * path1, const char * path2)
{
    Serial.printf("Renaming file %s to %s\n", path1, path2);
    if (fs.rename(path1, path2)) {
        Serial.println("File renamed");
    } else {
        Serial.println("Rename failed");
    }
}

void deleteFile(fs::FS &fs, const char * path){
    Serial.printf("Deleting file: %s\n", path);
    if(fs.remove(path)){
        Serial.println("File deleted");
    } else {
        Serial.println("Delete failed");
    }
}

void testFileIO(fs::FS &fs, const char * path)
{
    File file = fs.open(path);
    static uint8_t buf[512];
    size_t len = 0;
    uint32_t start = millis();
    uint32_t end = start;
    if(file){
        len = file.size();
        size_t flen = len;
        start = millis();
        while(len){
            size_t toRead = len;
            if(toRead > 512){
                toRead = 512;
            }
            file.read(buf, toRead);
            len -= toRead;
        }
        end = millis() - start;
        Serial.printf("%u bytes read for %u ms\n", flen, end);
        file.close();
    } else {
        Serial.println("Failed to open file for reading");
    }


    file = fs.open(path, FILE_WRITE);
    if(!file){
        Serial.println("Failed to open file for writing");
        return;
    }

    size_t i;
    start = millis();
    for(i=0; i<2048; i++){
        file.write(buf, 512);
    }
    end = millis() - start;
    Serial.printf("%u bytes written for %u ms\n", 2048 * 512, end);
    file.close();
}

int readFileText(fs::FS &fs, char *path, char *p)
{
    File file = fs.open( path);
    size_t len = 0;
    uint32_t start = millis();
    uint32_t end = start;
    if(file){
        len = file.size();
        size_t flen = len;
        start = millis();
        while(len){
            size_t toRead = len;
            if(toRead > 512){
                toRead = 512;
            }
            file.read( (uint8_t*)p, toRead);
            p += toRead;
            len -= toRead;
        }
        end = millis() - start;
        Serial.printf("%u bytes read for %u ms\n", flen, end);
        file.close();
        return 1;
    } else {
        Serial.println("Failed to open file for reading");
        return 0;
    }
}

/*
int	load_favicon( fs::FS &fs, char *path, char *p)
{
	int	len = 0;
	
    Serial.printf("Reading file: %s\n", path);

    File file = fs.open( path);
    if(!file){
        Serial.println("Failed to open file for reading");
        return 0;
    }
    len = file.size();
    if( len < sizeof( favicon)){
	    if( file.read( (uint8_t*)p, len)){
	        Serial.println("File read");
	    } else {
	        Serial.println("Read failed");
	    }
    } else {
    	len = 0;
    }
    file.close();
    return len;
}

*/

void	save_bin( fs::FS &fs, char *path, char *p, int size)
{
    Serial.printf("Writing file: %s\n", path);

    File file = fs.open( path, FILE_WRITE);
    if(!file){
        Serial.println("Failed to open file for writing");
        return;
    }
    if( file.write( (uint8_t*)p, size)){
        Serial.println("File written");
    } else {
        Serial.println("Write failed");
    }
    file.close();
}

void	save_jpeg( char *p, int size)
{
	char	fname[ 32];
	char	ddate[ 16];
	
	if( sdc_ok){
		sprintf( ddate, "/%02d%02d%02d", now.year, now.month, now.date);
	    createDir(SD, ddate);
		sprintf( fname, "%s/%02d%02d%02d.jpg", ddate, now.hour, now.minute, now.sec);
		save_bin( SD, fname, p, size);
		Serial.println("save picture");
	} else {
		Serial.println("reject picture");
	}
}

int sdc_init( void)
{
	int	spi_ok;
	
#ifdef	CONFIG_IDF_TARGET_ESP32S3
 	vspi = new SPIClass( VSPI);
 	if( vspi != NULL){
		vspi->begin( VSPI_SCK, VSPI_MISO, VSPI_MOSI, VSPI_SS);
		spi_ok = 1;
 	} else {
 		spi_ok = 0;
 	}
#else
	return 0;
#endif
	if( spi_ok){
		Serial.println(" VSPI Start");
	   if(!SD.begin( VSPI_SS, *vspi, SPI_FREQ)){
		   if(!SD.begin( VSPI_SS, *vspi, SPI_FREQ)){	// need 1 retry after power on
	     	   Serial.println("Card Mount Failed");
	        	return 0;
		   }
	    }
	    uint8_t cardType = SD.cardType();
	
	    if(cardType == CARD_NONE){
	        Serial.println("No SD card attached");
	        return 0;
	    }
	
	    Serial.print("SD Card Type: ");
	    if(cardType == CARD_MMC){
	        Serial.println("MMC");
	    } else if(cardType == CARD_SD){
	        Serial.println("SDSC");
	    } else if(cardType == CARD_SDHC){
	        Serial.println("SDHC");
	    } else {
	        Serial.println("UNKNOWN");
	    }
	
	    uint64_t cardSize = SD.cardSize() / (1024 * 1024);
	    Serial.printf("SD Card Size: %lluMB\n", cardSize);
	    Serial.printf("Total space: %lluMB\n", SD.totalBytes() / (1024 * 1024));
	    Serial.printf("Used space: %lluMB\n", SD.usedBytes() / (1024 * 1024));
	    LEDG_ON;
	    return 1;
	} else {
		return 0;
	}
}

void	i2c_init( void)
{
	Wire.begin( I2C_SDA, I2C_SCL);
}

void	init_DHT20( void)
{
	while( dht20.begin()){
	    Serial.println("Initialize sensor failed");
	    delay(1000);
	}
}

int	printLocalTime()
{
	
	if( getLocalTime( &tm_now)){
		Serial.println( &tm_now, "%A, %B %d %Y %H:%M:%S");
		now.year	= (tm_now.tm_year + 1900) % 100;
		now.month	= tm_now.tm_mon + 1;
		now.date	= tm_now.tm_mday;
		now.hour	= tm_now.tm_hour;
		now.minute	= tm_now.tm_min;
		now.sec		= tm_now.tm_sec;
		set_rv8263_data( now);
		return 1;
	} else {
	    Serial.println( "Failed to obtain time");
	    return 0;
	}
}

void	reset_data( void)
{
	int	i;
	
	for( i = 0 ; i < MAX_DATA; i++){
		temp[ i] = 0.0;
		humd[ i] = 0.0;
	}
}

void	reset_hist( void)
{
	int	i;
	
	for( i = 0 ; i < MAX_DATA; i++){
		hist_temp[ i] = 0.0;
		hist_humd[ i] = 0.0;
	}
}

char*	search_char ( char *p, char c)
{
	int	i;
	
	for( i = 0 ; i < 16; i++){
		if( *p++ == c){
			break;
		}
	}
	return p;	
}

void	load_data( void)
{
	char 	f[ 16];
	char* 	p;
	int		i;

	sprintf( f, "/%02d%02d%02d.csv", now.year, now.month, now.date);
	if( readFileText( SD, f, file_data)){
		p = file_data;
		p = search_char( p, '\n');
		for( i = 0; i < MAX_DATA; i++){
			temp[i] = atof( p);
			p = search_char( p, ',');
			humd[i] = atof( p);
			p = search_char( p, '\n');
		}
	} else {
		reset_data();
	}
}

void	save_data( void)
{
	char 	f[ 16];
	char* 	p;
	int		i;
	
	sprintf( f, "/%02d%02d%02d.csv", now.year, now.month, now.date);
	p = file_data;
	p += sprintf( p, "temperature,humidity\n");
	for( i = 0; i < MAX_DATA; i++){
		p += sprintf( p, "%4.1f,", temp[ i]);
		p += sprintf( p, "%4.1f\n", humd[ i]);
	}
	writeFile( SD, f, file_data);
}

void	set_ip_adr( char *p, char *pstr)
{
	*p++ = atoi( pstr);	pstr = search_char( pstr, '.');	
	*p++ = atoi( pstr);	pstr = search_char( pstr, '.');	
	*p++ = atoi( pstr);	pstr = search_char( pstr, '.');	
	*p++ = atoi( pstr);
}

void	set_config_string( int no, char *p)
{
	int	h,m;
	
	switch( no){
		case 1:	sprintf( para.ssid[0], "%s", p);	break;
		case 2:	sprintf( para.pawd[0], "%s", p);	break;
		case 3:	set_ip_adr( (char*)para.gway[0], p);	break;
		case 4:	set_ip_adr( (char*)para.ip[0], p);		break;
		case 5:	sprintf( para.ssid[1], "%s", p);	break;
		case 6:	sprintf( para.pawd[1], "%s", p);	break;
		case 7:	set_ip_adr( (char*)para.gway[1], p);	break;
		case 8:	set_ip_adr( (char*)para.ip[1], p);		break;
		case 9: sprintf( fname.ref_name[0], "%s", p);	break;
		case 10: sprintf( fname.ref_name[1], "%s", p);	break;
		case 11: sprintf( fname.ref_name[2], "%s", p);	break;
		case 12: sprintf( fname.ref_name[3], "%s", p);	break;
		case 13:
				h = atoi( p);	p = search_char( p, ':');	m = atoi( p);
				para.store_start = h * 60 + m;
				break;
		case 14:
				h = atoi( p);	p = search_char( p, ':');	m = atoi( p);
				para.store_end = h * 60 + m;
				break;
	}
	if( no){
		if(  no < 9){
			save_para();
		} else if(  no < 13){
			save_file_name();
		} else {
			save_para();
		}
	}
}

void	get_config_json( char *p)
{
    *p++ = '{';
	p += sprintf( p, "\"ssid1\":\"%s\"", para.ssid[ 0]);	p += sprintf( p, ",");
	p += sprintf( p, "\"ssid2\":\"%s\"", para.ssid[ 1]);	p += sprintf( p, ",");
	p += sprintf( p, "\"key1\":\"%s\"", para.pawd[ 0]);	p += sprintf( p, ",");
	p += sprintf( p, "\"key2\":\"%s\"", para.pawd[ 1]);	p += sprintf( p, ",");
	p += sprintf( p, "\"gway1\":\"%d.%d.%d.%d\"", para.gway[ 0][ 0], para.gway[ 0][1], para.gway[ 0][2], para.gway[ 0][3]);	p += sprintf( p, ",");
	p += sprintf( p, "\"gway2\":\"%d.%d.%d.%d\"", para.gway[ 1][ 0], para.gway[ 1][1], para.gway[ 1][2], para.gway[ 1][3]);	p += sprintf( p, ",");
	p += sprintf( p, "\"ip1\":\"%d.%d.%d.%d\"", para.ip[ 0][ 0], para.ip[ 0][1], para.ip[ 0][2], para.ip[ 0][3]);	p += sprintf( p, ",");
	p += sprintf( p, "\"ip2\":\"%d.%d.%d.%d\"", para.ip[ 1][ 0], para.ip[ 1][1], para.ip[ 1][2], para.ip[ 1][3]);	p += sprintf( p, ",");
	p += sprintf( p, "\"ref1\":\"%s\"", fname.ref_name[ 0]);	p += sprintf( p, ",");
	p += sprintf( p, "\"ref2\":\"%s\"", fname.ref_name[ 1]);	p += sprintf( p, ",");
	p += sprintf( p, "\"ref3\":\"%s\"", fname.ref_name[ 2]);	p += sprintf( p, ",");
	p += sprintf( p, "\"ref4\":\"%s\"", fname.ref_name[ 3]);	p += sprintf( p, ",");
	p += sprintf( p, "\"ss\":\"%02d:%02d\"", para.store_start/60,para.store_start % 60);	p += sprintf( p, ",");
	p += sprintf( p, "\"se\":\"%02d:%02d\"", para.store_end/60,para.store_end % 60);
    *p++ = '}';
    *p++ = 0;

}

void	load_hist_date( char *po)
{
	char 	f[ 16];
	char* 	p;
	int		i;

	sprintf( hist_name, "%s",po);
	sprintf( f, "/%s", hist_name);
	if( readFileText( SD, f, file_data)){
		p = file_data;
		p = search_char( p, '\n');
		for( i = 0; i < MAX_DATA; i++){
			hist_temp[i] = atof( p);
			p = search_char( p, ',');
			hist_humd[i] = atof( p);
			p = search_char( p, '\n');
		}
		hist_data_num = MAX_DATA;
	} else {
		sprintf( hist_name, "------.csv");
		reset_hist();
		hist_data_num = 0;
	}
}


void	load_history( int n)
{
	char 	f[ 16];
	char* 	p;
	int		i;

	sprintf( hist_name, "%s",fname.file_name[ fname.file_num+n]);
	sprintf( f, "/%s", hist_name);
	if( readFileText( SD, f, file_data)){
		p = file_data;
		p = search_char( p, '\n');
		for( i = 0; i < MAX_DATA; i++){
			hist_temp[i] = atof( p);
			p = search_char( p, ',');
			hist_humd[i] = atof( p);
			p = search_char( p, '\n');
		}
		hist_data_num = MAX_DATA;
	} else {
		sprintf( hist_name, "------.csv");
		reset_hist();
		hist_data_num = 0;
	}
}

void	regist_file_name( void)
{
	int	i, j;

	if( fname.today.date > 0){
		if( fname.file_num < 10){
			sprintf( fname.file_name[ fname.file_num++], "%02d%02d%02d.csv", fname.today.year, fname.today.month, fname.today.date);
		} else {
			for( i = 1; i < 10; i++){
				for( j = 0; j < 16; j++){
					fname.file_name[ i-1][ j] = fname.file_name[ i][ j];
				}
			}
			sprintf( fname.file_name[ fname.file_num-1], "%02d%02d%02d.csv", fname.today.year, fname.today.month, fname.today.date);
		}
	}
	fname.today = now;
	save_file_name();
}

void	printSerial( char *p)
{
	Serial.println( p);
}


void	init_time( void)
{

	int	i, c;

	for( c =0 ; c < 2; c++){
		WiFi.begin( para.ssid[ c], para.pawd[ c]);
   		Serial.print(para.ssid[ c]);
		for( i = 0 ; i < 10; i++){
	  		if( WiFi.status() != WL_CONNECTED) {
	    		delay(500);
	    		Serial.print(".");
			} else {
				Serial.println(" CONNECTED");
				configTime(gmtOffset_sec, daylightOffset_sec, ntpServer);
				time_ok = printLocalTime();
	    		Serial.println("local time get");
				WiFi.disconnect(true);
				WiFi.mode(WIFI_OFF);
				return;
			}
		}
   		Serial.println("skip");
	}
}

void setup( void)
{
	int	i;
	int	c;
	
	Serial.begin(115200);
	Serial.setDebugOutput(true);
	Serial.println();
	eeprom_ok = EEPROM.begin( EEP_SIZE);
	load_para();
	load_file_name();


	sd_save 	= 0;
	time_ok		= 0;
	eeprom_ok	= 0;
	sdc_ok		= 0;
	wifi_ok		= 0;
	data_num	= 0;
	reset_data();
	
	led_init( );
	i2c_init();
	rv8263_init();
	init_time();
	sprintf( time_string, "----/--/-- --:--:--");

	camera_config_t config;
	config.pin_d0 = Y2_GPIO_NUM;
	config.pin_d1 = Y3_GPIO_NUM;
	config.pin_d2 = Y4_GPIO_NUM;
	config.pin_d3 = Y5_GPIO_NUM;
	config.pin_d4 = Y6_GPIO_NUM;
	config.pin_d5 = Y7_GPIO_NUM;
	config.pin_d6 = Y8_GPIO_NUM;
	config.pin_d7 = Y9_GPIO_NUM;
	config.pin_xclk = XCLK_GPIO_NUM;
	config.pin_pclk = PCLK_GPIO_NUM;
	config.pin_vsync = VSYNC_GPIO_NUM;
	config.pin_href = HREF_GPIO_NUM;
	config.pin_sscb_sda = SIOD_GPIO_NUM;
	config.pin_sscb_scl = SIOC_GPIO_NUM;
	config.pin_pwdn = PWDN_GPIO_NUM;
	config.pin_reset = RESET_GPIO_NUM;
	config.xclk_freq_hz = 14000000;
	config.pixel_format = PIXFORMAT_JPEG;

	// if PSRAM IC present, init with UXGA resolution and higher JPEG quality
	//											for larger pre-allocated frame buffer.
	if( psramFound()){
		config.frame_size = FRAMESIZE_SVGA;	// FRAMESIZE_UXGA(1600x1200) or FRAMESIZE_SVGA(800x600)
		config.jpeg_quality = 12;			// 10;
		config.fb_count = 1;				// 2;
	} else {
		config.frame_size = FRAMESIZE_SVGA;
		config.jpeg_quality = 12;
		config.fb_count = 1;
	}

	// camera init
	esp_err_t err = esp_camera_init(&config);
	if (err != ESP_OK) {
		Serial.printf("Camera init failed with error 0x%x", err);
		return;
	}

	sensor_t * s = esp_camera_sensor_get();
	// initial sensors are flipped vertically and colors are a bit saturated
	if( s->id.PID == OV2640_PID){
		Serial.println("Camera OV2640 detected");
	}

#if defined(CAMERA_MODEL_M5STACK_WIDE) || defined(CAMERA_MODEL_M5STACK_ESP32CAM)
	s->set_vflip(s, 1);
	s->set_hmirror(s, 1);
#endif

	for( c = 0 ; c < 2; c++){
		WiFi.config( para.ip[ c], para.gway[ c],subnet);
		WiFi.begin( para.ssid[ c], para.pawd[ c]);

		Serial.print( para.ssid[ c]);
		for( i = 0 ; i < 10; i++){
			if (WiFi.status() != WL_CONNECTED) {
				delay(500);
				Serial.print(".");
			} else {
				Serial.println("");
				Serial.println("WiFi connected");
				wifi_ok = 1;
				break;
			}
		}
		if( i < 10) break;
		Serial.println("skip");
	}
	LEDR_OFF;

	if( wifi_ok){
		startCameraServer();
		Serial.print("Camera Ready! Use 'http://");
		Serial.print(WiFi.localIP());
		Serial.println("' to connect");
	}
	sdc_ok = sdc_init();
	if( sdc_ok){
//		favicon_len = load_favicon( SD, "/favicon.ico", favicon);
		load_data();
	}
	init_DHT20();
	fname.today = now;
	save_file_name();
	sprintf( today_name, "%02d%02d%02d.csv", now.year, now.month, now.date);
}

int	old_data_no;
int	old_date;

void	clear_data( int num)
{
	int	i;

	for( i = num; i < MAX_DATA; i++){
		temp[i] = 0.0;
		humd[i] = 0.0;
	}
}

void loop()
{
	int	n;
	
 	delay(100);
 	temperature = dht20.getTemperature();
 	humidity	= dht20.getHumidity() * 100.0;
  	get_rv8263_data( &now);
 	sprintf( time_string, "%4d/%2d/%2d %2d:%2d:%2d", now.year, now.month, now.date,now.hour, now.minute, now.sec);
 	n = now.hour * 60 + now.minute;
 	n = n % MAX_DATA;
 	data_num = n;
 	if( n != old_data_no){			// every minute process
 		if( now.date != fname.today.date){	// 
 			clear_data( n);			// clear future data
 			regist_file_name();
 		}
	 	temp[ n] = temperature;
	 	humd[ n] = humidity;
 		old_data_no = n;
 		if( sdc_ok){
	 		save_data();
	 		sd_save = 1;
	 		if(( n >= para.store_start) && ( n <= para.store_end)){
		 		camera_capture();
	 		}
 		}
 	}
}
